/*
 ****************************************************************
 *								*
 *			lockf.c					*
 *								*
 *	Algoritmos dos LOCKs de regies de arquivos		*
 *								*
 *	Verso	3.0.0, de 24.11.94				*
 *		3.0.0, de 24.11.94				*
 *								*
 *	Mdulo: Ncleo						*
 *		NCLEO do TROPIX para PC			*
 *								*
 *	TROPIX: Sistema Operacional Tempo-Real Multiprocessado	*
 *		Copyright  1999 NCE/UFRJ - tecle "man licena"	*
 *								*
 ****************************************************************
 */

#include "../h/common.h"
#include "../h/sync.h"
#include "../h/scb.h"
#include "../h/region.h"

#include "../h/inode.h"
#include "../h/lockf.h"
#include "../h/uerror.h"
#include "../h/signal.h"
#include "../h/uproc.h"

#include "../h/extern.h"
#include "../h/proto.h"

/*
 ****************************************************************
 *	Variveis e definies globais				*
 ****************************************************************
 */
#define	PLOCKF	PPIPE

/*
 ****************************************************************
 *	Variveis locais					*
 ****************************************************************
 */
entry LOCKF	*lffreelist;	/* Lista livre */
entry LOCK	lffreelock;	/* Semforo da lista livre */

/*
 ****************************************************************
 *	Testa se pode acessar uma regio			*
 ****************************************************************
 */
int
lfaccess (INODE *ip, off_t begin_off, off_t end_off, int flag)
{
	LOCKF		*lp;

	/*
	 *	Todos as funes deste arquivo
	 *	recebem e devolvem o INODE locked 
	 *
	 *	"flag" indica qual o procedimento a tomar
	 *	em caso da rea estar trancada:
	 *	   'W' => dorme aguardando liberao
	 *	   'T' => retorna imediatamente indicando erro
	 */
    again:
	for (lp = ip->i_lockf; lp != NOLOCKF; lp = lp->k_next)
	{
		if (lp->k_uproc == u.u_myself)
			continue;

		if (end_off > lp->k_begin && begin_off < lp->k_end)
			goto found;
	}

	return (0);

	/*
	 *	Foi encontrada uma regio de litgio
	 */
    found:
	switch (flag)
	{
	    default:
		printf ("%g: Indicador errado\n");
	    case 'T':
		u.u_error = EAGAIN;
		return (-1);

	    case 'W':
		break;

	}	/* end switch */

	/*
	 *	A regio est trancada, e devemos esperar
	 */
	if (lfdeadlock (lp) < 0)
		return (-1);

	SLEEPFREE (&ip->i_lock);

	EVENTWAIT (&lp->k_free, PLOCKF);

	SLEEPLOCK (&ip->i_lock, PINODE);

	goto again;

}	/* end lfaccess */

/*
 ****************************************************************
 *	Cria um LOCK						*
 ****************************************************************
 */
int
lflock (INODE *ip, off_t begin_off, off_t end_off, int flag)
{
	LOCKF		*lp, *ap, *np;
	LOCKF		*gp;

	/*
	 *	Certifica-se de que outro processo j no possui a regio
	 */
	if (lfaccess (ip, begin_off, end_off, flag) < 0)
		return (-1);

	/*
	 *	Obtm um bloco para a lista (neste momento  mais seguro)
	 */
	if ((gp = lfget ()) == NOLOCKF)
		{ u.u_error = EDEADLK; return (-1); }

	/*
	 *	Faz o merge com outras reas possudas pelo processo 
	 */
	for
	(	ap = NOLOCKF, lp = ip->i_lockf;
		lp != NOLOCKF;
		ap = lp, lp = np
	)
	{
		np = lp->k_next;

		if (end_off < lp->k_begin || begin_off > lp->k_end)
			continue;

		if (lp->k_uproc != u.u_myself)
		{
			if (end_off != lp->k_begin && begin_off != lp->k_end)
				printf ("%g: Confuso com LOCKs alheios\n");

			continue;
		}

		/*
		 *	Calcula a unio
		 */
		if (begin_off > lp->k_begin)
			begin_off = lp->k_begin;

		if (end_off   < lp->k_end)
			end_off   = lp->k_end;

		/*
		 *	Libera este LOCK; processos esperando
		 *	por ele analisaro a lista novamente
		 */
		EVENTDONE (&lp->k_free);

		lfput (lp);

		if (ap == NOLOCKF)
			ip->i_lockf = np;
		else
			ap->k_next  = np;

		lp = ap;

	}	/* end for */

	/*
	 *	Cria o novo bloco da lista
	 */
	lp = gp;

	EVENTCLEAR (&lp->k_free);
	lp->k_uproc = u.u_myself;
	lp->k_begin = begin_off;
	lp->k_end   = end_off;

	lp->k_next  = ip->i_lockf;
	ip->i_lockf = lp;

	return (0);

}	/* end lflock */

/*
 ****************************************************************
 *	Libera um LOCK						*
 ****************************************************************
 */
int
lffree (INODE *ip, off_t begin_off, off_t end_off)
{
	LOCKF		*lp, *ap, *np;
	LOCKF		*gp;
	off_t		k_end;

	/*
	 *	Obtm um bloco para a lista j prevendo um possvel buraco
	 *	(neste momento  mais seguro)
	 */
	if ((gp = lfget ()) == NOLOCKF)
		{ u.u_error = EDEADLK; return (-1); }

	/*
	 *	Faz o merge com outras reas possudas pelo processo 
	 */
	for
	(	ap = NOLOCKF, lp = ip->i_lockf;
		lp != NOLOCKF;
		ap = lp, lp = np
	)
	{
		np = lp->k_next;

		if (lp->k_uproc != u.u_myself)
			continue;

		if (end_off <= lp->k_begin || begin_off >= lp->k_end)
			continue;

		/*
		 *	Libera este LOCK; processos esperando
		 *	por ele analisaro a lista novamente
		 */
		EVENTDONE (&lp->k_free);

		/*
		 *	Analisa o que resta da regio
		 */
		if (begin_off <= lp->k_begin)
		{
			if (end_off >= lp->k_end)
			{
				/*
				 *	Nada resta
				 */
				lfput (lp);

				if (ap == NOLOCKF)
					ip->i_lockf = np;
				else
					ap->k_next  = np;

				lp = ap;
			}
			else
			{
				/*
				 *	Resta um pedao em cima
				 */
				lp->k_begin = end_off;
			}
		}
		else
		{
			k_end = lp->k_end;
			lp->k_end = begin_off;

			/*
			 *	Verifica se restou um pedao em cima
			 */
			if (end_off < k_end)
			{
				/*
				 *	A rea ser transformada em duas
				 */
				if (gp == NOLOCKF)
					panic ("lffree: 2 transformaes?");

				EVENTCLEAR (&gp->k_free);
				gp->k_uproc  = u.u_myself;
				gp->k_begin = end_off;
				gp->k_end   = k_end;

				gp->k_next  = np;
				lp->k_next  = gp;

				ap = lp;
				lp = gp;

				gp = NOLOCKF;

			}	/* end if transformao de regio? */ 
				
		}	/* end if interceptou em baixo? */  

	}	/* end for */

	/*
	 *	Se no precisou do novo bloco, devolve
	 */
	if (gp != NOLOCKF)
		lfput (gp);

	return (0);

}	/* end lffree */

/*
 ****************************************************************
 *	Verifica se h o perigo de DEADLOCK			*
 ****************************************************************
 */
int
lfdeadlock (LOCKF *lp)
{
	void		*sema;

	/*
	 *	Se o processo dono da regio no est dormindo, no h perigo
	 */
    again:
	sema = lp->k_uproc->u_sema;

	if (sema == (void *)NULL)
		return (0);

	/*
	 *	O processo est dormindo; tenta identificar o semforo
	 *	Se estiver dormindo por outras causas no h perigo.
	 */
	if ((int)sema < (int)scb.y_lockf || (int)sema >= (int)scb.y_endlockf)
		return (0);

	/*
	 *	Est dormindo em virtude de uma outra regio; tenta achar o dono
	 *	dela; repare que o semforo  o primeiro membro da entrada
	 */
	lp = (LOCKF *)sema;

	if (((int)lp - (int)scb.y_lockf) % sizeof (LOCKF))
		{ printf ("%g: Endereo de semforo invlido\n"); goto bad; }

	/*
	 *	Verifica o dono da regio; se for este processo  DEADLOCK
	 */
	if (lp->k_uproc != u.u_myself)
		goto again;

	/*
	 *	Em caso de erro, ...
	 */
    bad:
	u.u_error = EDEADLK;
	return (-1);

}	/* end lfdeadlock */

/*
 ****************************************************************
 *	Libera todos os LOCKs possudos por este processo	*
 ****************************************************************
 */
int
lfclosefree (INODE *ip)
{
	LOCKF		*ap, *lp, *np;

	/*
	 *	Percorre a lista, liberando todas as regies deste processo
	 */
	for
	(	ap = NOLOCKF, lp = ip->i_lockf;
		lp != NOLOCKF;
		ap = lp, lp = np
	)
	{
		np = lp->k_next;

		if (lp->k_uproc != u.u_myself)
			continue;
		
		/*
		 *	Encontrou uma regio deste processo
		 */
		EVENTDONE (&lp->k_free);

		lfput (lp);

		if (ap == NOLOCKF)
			ip->i_lockf = np;
		else
			ap->k_next  = np;

		lp = ap;

	}	/* end for */

	return (0);

}	/* end lfclosefree */

/*
 ****************************************************************
 *	Obtm uma estrutura LOCKF				*
 ****************************************************************
 */
LOCKF *
lfget (void)
{
	LOCKF		*lp;

	SPINLOCK (&lffreelock);

	if ((lp = lffreelist) != NOLOCKF)
		lffreelist = lp->k_next;

	SPINFREE (&lffreelock);

	return (lp);

}	/* end lfget */

/*
 ****************************************************************
 *	Libera uma estrutura LOCKF				*
 ****************************************************************
 */
void
lfput (LOCKF *lp)
{
	SPINLOCK (&lffreelock);

	lp->k_next = lffreelist;
	lffreelist = lp;

	SPINFREE (&lffreelock);

}	/* end lfput */

/*
 ****************************************************************
 *	Inicializa a lista de LOCKs				*
 ****************************************************************
 */
void
lfinit (void)
{
	LOCKF		*lp, *np;

	np = NOLOCKF;

	for (lp = scb.y_endlockf - 1; lp >= scb.y_lockf; lp--)
		{ lp->k_next = np; np = lp; }

	lffreelist = np;

}	/* end lfinit */
