/*
 ****************************************************************
 *								*
 *			<sys/region.h>				*
 *								*
 *	Tabelas de regies de memria				*
 *								*
 *	Verso	3.0.0, de 07.09.94				*
 *		4.6.0, de 22.08.04				*
 *								*
 *	Mdulo: Ncleo						*
 *		NCLEO do TROPIX para PC			*
 *		/usr/include/sys				*
 *								*
 *	TROPIX: Sistema Operacional Tempo-Real Multiprocessado	*
 *		Copyright  2004 NCE/UFRJ - tecle "man licena"	*
 *								*
 ****************************************************************
 */

#define	REGION_H		/* Para definir os prottipos */

/*
 ******	Formato das estruturas **********************************
 */
struct	regionl
{
	char	rgl_type;	/* Tipo da regio */
	char	rgl_prot;	/* Tipo de proteo da regio */

	pg_t	rgl_vaddr;	/* Endereo virtual em pginas na memria */
	REGIONG	*rgl_regiong;	/* Ponteiro para a regio global */
};

struct	regiong
{
	char	rgg_name[16];	/* Nome da regio */
	char	rgg_type;	/* Tipo da regio */
	char	rgg_flags;	/* Caractersticas da regio */
	LOCK	rgg_regionlock;	/* SPINLOCK de uso exclusivo da entrada */
	char	rgg_pgtb_sz;	/* Tamanho da tabela de pginas da regio */

	pg_t	rgg_size;	/* Tamanho da regio em pginas */
	pg_t	rgg_paddr;	/* Endereo real em pginas na memria */

	pg_t	rgg_pgtb;	/* Tabela de pginas da regio */
	int	rgg_count;	/* No. de proc. usando esta regio (procreg) */
	REGIONG	*rgg_next;	/* Prx. reg. assoc. mesmo i-node (shmemreg) */
};

/*
 ******	Cdigo das regies **************************************
 */
enum
{
	RG_NONE,	/* Regio inexistente */
	RG_TEXT,	/* Regio de texto compartilhado */
	RG_DATA,	/* Regio de data (compart. entre threads) */
	RG_STACK,	/* Regio de stack nunca compartilhada */
	RG_PHYS,	/* Regio de PHYS */
	RG_SHMEM	/* Regio de memria compartilhada */
};

#define RG_NAMES { "NONE", "TEXT", "DATA", "STACK", "PHYS", "SHMEM" }

/*
 ******	Indicadores para as operaes com regies ***************
 */
enum
{
	RG_CLR	   = 1,		/* Regio deve ser zerada */
	RG_NOMMU   = 2,		/* No carrega/descarrega MMU */

	RG_PHYSDUP = 0,		/* Regio deve ser duplicada fsicamente */
	RG_LOGDUP  = 4		/* Regio deve ser duplicada logicamente */
};

/*
 ******	Indicadores para as regies globais *********************
 */
enum
{
	RGG_METX   = 0x04	/* Manter cpia do processo em memria */
};

#define	RGG_FLAG_BITS	"\x03" "METX"

/*
 ******	Indicadores de criao de processos *********************
 */
enum { FORK, THREAD };
